package org.prss.contentdepot.cdedge.util;

import com.pi4j.io.spi.Spi;
import com.pi4j.io.spi.SpiBase;
import com.pi4j.io.spi.SpiConfig;
import com.pi4j.io.spi.SpiProvider;
import com.sun.jna.Memory;
import com.sun.jna.Native;
import com.sun.jna.Structure;
import com.sun.jna.ptr.IntByReference;
import static org.prss.contentdepot.cdedge.util.Pi4JLibC.*;

/**
 * SPI implementation that uses JNA bindings to the Linux SPI device (i.e. /dev/spidev0.0). Only supports writing but
 * it works to drive an SSD1306 OLED display.
 *
 * @author mpilone
 * @see
 * <a href="https://github.com/sckulkarni246/ke-rpi-samples/blob/main/spi-c-ioctl/oled_functions.c#L157">spi-c-ioctl</a>
 * @see <a href="https://github.com/torvalds/linux/blob/master/include/uapi/linux/spi/spidev.h">spidev.h</a>
 * @see <a href="https://github.com/torvalds/linux/blob/master/include/uapi/linux/spi/spi.h">spi.h</a>
 * @since 10/3/24.
 */
public class JnaSpi extends SpiBase implements Spi {

    ///////////////////////////////////
    // From spidev.h
    private final static byte SPI_IOC_MAGIC = 'k';
    private final static byte SIZE_OF_BYTE = 1;
    private final static byte SIZE_OF_INT = 4;

//#define SPI_MSGSIZE(N) \
//            ((((N)*(sizeof (struct spi_ioc_transfer))) < (1 << _IOC_SIZEBITS)) \
//            ? ((N)*(sizeof (struct spi_ioc_transfer))) : 0)
//            #define SPI_IOC_MESSAGE(N) _IOW(SPI_IOC_MAGIC, 0, char[SPI_MSGSIZE(N)])
    private static int SPI_IOC_MESSAGE(int n) {

        // Even though we will pass the structure to ioctl as a pointer, the command needs to know
        // the actual size of the structure (i.e. sizeof). Therefore, we use the ByValue interface
        // when getting the struct size.
        int structSize = Native.getNativeSize(spi_ioc_transfer.ByValue.class);
        int msgSize = ((((n)*(structSize)) < (1 << _IOC_SIZEBITS))
                ? ((n)*(structSize)) : 0);

        return _IOC(_IOC_WRITE, SPI_IOC_MAGIC, (byte)0, msgSize);
    }

    /* Read / Write of SPI mode (SPI_MODE_0..SPI_MODE_3) (limited to 8 bits) */
    private final static int SPI_IOC_RD_MODE = _IOC(_IOC_READ, SPI_IOC_MAGIC, (byte)1, SIZE_OF_BYTE);
    private final static int SPI_IOC_WR_MODE = _IOC(_IOC_WRITE, SPI_IOC_MAGIC, (byte)1, SIZE_OF_BYTE);

    /* Read / Write SPI bit justification */
    private final static int SPI_IOC_RD_LSB_FIRST = _IOC(_IOC_READ, SPI_IOC_MAGIC, (byte)2, SIZE_OF_BYTE);
    private final static int SPI_IOC_WR_LSB_FIRST = _IOC(_IOC_WRITE, SPI_IOC_MAGIC, (byte)2, SIZE_OF_BYTE);

    /* Read / Write SPI device word length (1..N) */
    private final static int SPI_IOC_RD_BITS_PER_WORD = _IOC(_IOC_READ, SPI_IOC_MAGIC, (byte)3, SIZE_OF_BYTE);
    private final static int SPI_IOC_WR_BITS_PER_WORD = _IOC(_IOC_WRITE, SPI_IOC_MAGIC, (byte)3, SIZE_OF_BYTE);

    /* Read / Write SPI device default max speed hz */
    private final static int SPI_IOC_RD_MAX_SPEED_HZ = _IOC(_IOC_READ, SPI_IOC_MAGIC, (byte)4, SIZE_OF_INT);
    private final static int SPI_IOC_WR_MAX_SPEED_HZ = _IOC(_IOC_WRITE, SPI_IOC_MAGIC, (byte)4, SIZE_OF_INT);

    /* Read / Write of the SPI mode field */
    private final static int SPI_IOC_RD_MODE32 = _IOC(_IOC_READ, SPI_IOC_MAGIC, (byte)5, SIZE_OF_INT);
    private final static int SPI_IOC_WR_MODE32 = _IOC(_IOC_WRITE, SPI_IOC_MAGIC, (byte)5, SIZE_OF_INT);

    @Structure.FieldOrder({"tx_buf", "rx_buf",
            "len", "speed_hz",
            "delay_usecs", "bits_per_word", "cs_change", "tx_nbits", "rx_nbits", "word_delay_usecs", "pad"})
    public static class spi_ioc_transfer extends Structure {
        public long tx_buf;
        public long rx_buf;

        public int		len;
        public int		speed_hz;

        public short	delay_usecs;
        public byte		bits_per_word;
        public byte		cs_change;
        public byte		tx_nbits;
        public byte		rx_nbits;
        public byte		word_delay_usecs;
        public byte		pad;

        public static class ByValue extends spi_ioc_transfer implements Structure.ByValue {}
    }

    ///////////////////////////////////
    // spi.h
    private final byte SPI_MODE_0	=	(0|0);

    private final static String SPI_DEVICE = "/dev/spidev0.0";
    private final Pi4JLibC libc = Pi4JLibC.INSTANCE;
    private int fd;

    public JnaSpi(SpiProvider provider, SpiConfig config) {
        super(provider, config);
    }

    @Override
    public void open() {
        super.open();

        fd = libc.open(SPI_DEVICE, Pi4JLibC.O_RDWR);
        if (fd < 0) {
            throw new RuntimeException("Failed to open SPI device " + SPI_DEVICE);
        }

        IntByReference intPtr = new IntByReference();
        int ret = libc.ioctl(fd, SPI_IOC_RD_MODE32, intPtr);
        if(ret != 0) {
            libc.close(fd);
            throw new RuntimeException("Could not read SPI mode.");
        }

        intPtr.setValue(intPtr.getValue() | SPI_MODE_0);

        ret = libc.ioctl(fd, SPI_IOC_WR_MODE32, intPtr);
        if(ret != 0) {
            libc.close(fd);
            throw new RuntimeException("Could not write SPI mode..");
        }

        ret = libc.ioctl(fd, SPI_IOC_RD_MAX_SPEED_HZ, intPtr);
        if(ret != 0) {
            libc.close(fd);
            throw new RuntimeException("Could not read the SPI max speed.");
        }

        intPtr.setValue(100_000);
        ret = libc.ioctl(fd, SPI_IOC_WR_MAX_SPEED_HZ, intPtr);
        if(ret != 0) {
            libc.close(fd);
            throw new RuntimeException("Could not write the SPI max speed.");
        }
    }

    @Override
    public void close() {
        libc.close(fd);

        super.close();
    }

    @Override
    public int transfer(byte[] write, int writeOffset, byte[] read, int readOffset, int numberOfBytes) {
        // Not implemented, I don't have a device to read from. But it should use the same
        // spi_ioc_transfer structure as write(...).
        return 0;
    }

    @Override
    public int read() {
        // Not implemented, I don't have a device to read from. But it should use the same
        // spi_ioc_transfer structure as write(...).
        return 0;
    }

    @Override
    public int read(byte[] buffer, int offset, int length) {
        // Not implemented, I don't have a device to read from. But it should use the same
        // spi_ioc_transfer structure as write(...).
        return 0;
    }

    @Override
    public int write(byte b) {
        return write(new byte[] {b}, 0, 1);
    }

    @Override
    public int write(byte[] data, int offset, int length) {
        // We could reuse this buffer for all requests.
        PeerAccessibleMemory buf = new PeerAccessibleMemory(length);
        buf.write(0, data, offset, length);

        spi_ioc_transfer transfer = new spi_ioc_transfer();
        transfer.tx_buf = buf.getPeer();
        transfer.rx_buf = 0;
        transfer.bits_per_word = 0;
        transfer.speed_hz = 100_000;
        transfer.delay_usecs = 0;
        transfer.len = length;

        int ret = libc.ioctl(fd, SPI_IOC_MESSAGE(1), transfer);
        if (ret < 0) {
            System.err.println("Could not write SPI message: " + ret);
            System.err.println("Last error: " + Native.getLastError());
        }

        buf.close();

        return length;
    }

    /**
     * Extension of Memory to allow direct access to the native peer pointer. This is required because
     * the {@link spi_ioc_transfer} structure uses a long for the tx_buf and rx_buf fields but a
     * native pointer is normally only 32 bits. We put the peer in a long in the struct to maintain
     * proper struct alignment.
     */
    private static class PeerAccessibleMemory extends Memory {
        public PeerAccessibleMemory(long size) {
            super(size);
        }

        long getPeer() {
            return peer;
        }
    }
}
